"""
Licensed under the Apache License, Version 2.0 (the "License");
You may not use this file except in compliance with the License.
You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and* limitations under the License.*
"""
import argparse
import configparser
import getpass
import os
from typing import List

import keyring

from . import errors, ui

LAZY_VPN_KEYCHAIN_SERVICE_NAME = 'lazyvpn'
DEFAULT_VPN_REGION = [
    'ZillowGroup Seattle',
    'ZillowGroup Denver',
    'ZillowGroup Irvine',
    'ZillowGroup Cincinnati',
    'ZillowGroup Lincoln',
    'ZillowGroup New York',
    'ZillowGroup San Francisco',
]


class Config(object):
    """
       The Config Class gets the CLI arguments, writes out the okta config file,
       gets and returns username and password and the Okta API key.

       A lot of this code is modified from https://github.com/nimbusscale/okta_aws_login
       under the MIT license.
    """

    def __init__(self, gac_ui, create_config=True):
        """
        :type gac_ui: ui.UserInterface
        """
        self.ui = gac_ui
        self.FILE_ROOT = self.ui.HOME
        self.LAZYVPN_CONFIG = self.ui.environ.get('LAZYVPN_CONFIG', os.path.join(self.FILE_ROOT, '.lazyvpn_config'))

        self.conf_profile = 'DEFAULT'
        self.action_configure = False
        self.username = None
        self.company_vpn_region = None
        self.up = False
        self.down = False

        if create_config and not os.path.isfile(self.LAZYVPN_CONFIG):
            self.ui.notify('No lazy-vpn configuration file found, starting first-time configuration...')
            self.update_config_file()

    def get_args(self):
        """Get the CLI args"""
        parser = argparse.ArgumentParser(
            description="Create connection to VPN via Okta"
        )
        parser.add_argument('--profile', '-p', help='If set, the specified configuration profile will be used instead of the default.')
        parser.add_argument('--action-configure', '--configure', '-c', action='store_true', help="If set, will prompt user for configuration parameters and then exit.")
        parser.add_argument('--up', '-u', action='store_true', help='connect to vpn')
        parser.add_argument('--down', '-d', action='store_true', help='disconnect from vpn')
        args = parser.parse_args(self.ui.args)
        self.action_configure = args.action_configure
        self.conf_profile = args.profile or 'DEFAULT'
        if not self.action_configure and args.up == args.down:
            raise ValueError("Can't have both be UP and DOWN the same value")
        return args

    def _handle_config(self, config, profile_config, include_inherits=True):
        if "inherits" in profile_config.keys() and include_inherits:
            self.ui.message("Using inherited config: " + profile_config["inherits"])
            if profile_config["inherits"] not in config:
                raise errors.LazyVpnError(self.conf_profile + " inherits from " + profile_config["inherits"] + ", but could not find " + profile_config["inherits"])
            combined_config = {
                **self._handle_config(config, dict(config[profile_config["inherits"]])),
                **profile_config,
            }
            del combined_config["inherits"]
            return combined_config
        else:
            return profile_config

    def get_config_dict(self, include_inherits=True):
        """returns the conf dict from the okta config file"""
        # Check to see if config file exists, if not complain and exit
        # If config file does exist return config dict from file
        if os.path.isfile(self.LAZYVPN_CONFIG):
            config = configparser.ConfigParser()
            config.read(self.LAZYVPN_CONFIG)
            try:
                profile_config = dict(config[self.conf_profile])
                self.fail_if_profile_not_found(profile_config, self.conf_profile, config.default_section)
                return self._handle_config(config, profile_config, include_inherits)
            except KeyError:
                if self.action_configure:
                    return {}
                raise errors.LazyVpnError(
                    'Configuration profile not found! Use the --configure flag to generate the profile.')
            return config
        raise errors.LazyVpnError('Configuration file not found! Use the --configure flag to generate file.')

    def update_config_file(self):
        """
           Prompts user for config details for the okta_aws_login tool.
           Either updates existing config file or creates new one.
           Config Options:
                okta_username = Okta username
        """
        config = configparser.ConfigParser()
        if self.action_configure:
            self.conf_profile = self._get_conf_profile_name(self.conf_profile)

        defaults = {
            'okta_username': '',
            'company_vpn_region': DEFAULT_VPN_REGION[0]
        }

        # See if a config file already exists.
        # If so, use current values as defaults
        if os.path.isfile(self.LAZYVPN_CONFIG):
            config.read(self.LAZYVPN_CONFIG)
            if self.conf_profile in config:
                profile = config[self.conf_profile]
                for default in defaults:
                    defaults[default] = profile.get(default, defaults[default])

        # Prompt user for config details and store in config_dict
        config_dict = defaults
        config_dict['okta_username'] = self._get_okta_username(defaults['okta_username'])
        self._set_okta_password(config_dict['okta_username'])
        config_dict['company_vpn_region'] = self._get_company_vpn_region(DEFAULT_VPN_REGION)
        self.write_config_file(config_dict)

    def write_config_file(self, config_dict):
        config = configparser.ConfigParser()
        config.read(self.LAZYVPN_CONFIG)
        config[self.conf_profile] = config_dict
        # write out the conf file
        with open(self.LAZYVPN_CONFIG, 'w') as configfile:
            config.write(configfile)

    def _get_user_input_for_password(self, username, message):
        for x in range(0, 5):
            passwd_prompt = f"{message}: "
            password = getpass.getpass(prompt=passwd_prompt)
            if len(password) > 0:
                break
        # If the OS supports a keyring, offer to save the password
        if self.ui.input("Do you want to save this password in the keyring? (y/N) ") == 'y':
            try:
                keyring.set_password(
                    LAZY_VPN_KEYCHAIN_SERVICE_NAME, username, password)
                self.ui.info("Password for {} saved in keyring.".format(username))
            except RuntimeError as err:
                self.ui.warning("Failed to save password in keyring: " + str(err))

    def _get_user_input_with_default_value(self, message, default=None, hidden=False):
        """formats message to include default and then prompts user for input
        via keyboard with message. Returns user's input or if user doesn't
        enter input will return the default."""
        if default and default != '':
            prompt_message = message + " [{}]: ".format(default)
        else:
            prompt_message = message + ': '

        # print the prompt with print() rather than input() as input prompts on stderr
        user_input = self.ui.input(prompt_message, hidden)
        if user_input:
            return user_input
        return default

    def _get_user_input_with_predefined_list(self, message, default=[], hidden=False):
        """formats message to include default and then prompts user for input
        via keyboard with message. Returns user's input or if user doesn't
        enter input will return the default."""
        if default and len(default) > 0:
            prompt_message = message + ":\n"
            for index, entry in enumerate(default):
                prompt_message += f"{index} - {entry}\n"
        else:
            prompt_message = message + ': '

        # print the prompt with print() rather than input() as input prompts on stderr
        user_input = self.ui.input(prompt_message, hidden)
        if user_input:
            return default[int(user_input)]
        return default[0]

    def clean_up(self):
        """ clean up secret stuff"""
        del self.username

    def fail_if_profile_not_found(self, profile_config, conf_profile, default_section):
        """
        When a users profile does not have a profile named 'DEFAULT' configparser fails to throw
        an exception. This will raise an exception that handles this case and provide better messaging
        to the user why the failure occurred.
        Ensure that whichever profile is set as the default exists in the end users okta config
        """
        if not profile_config and conf_profile == default_section:
            raise errors.LazyVpnError(
                'DEFAULT profile is missing! This is profile is required when not using --profile')

    def _get_conf_profile_name(self, default_entry):
        """Get and validate configuration profile name. [Optional]"""
        ui.default.message(
            "If you'd like to assign the Okta configuration to a specific profile\n"
            "instead of to the default profile, specify the name of the profile.\n"
            "This is optional.")
        conf_profile = self._get_user_input_with_default_value(
            "Okta Configuration Profile Name", default_entry)
        return conf_profile

    def _get_okta_username(self, default_entry):
        """Get and validate okta username. [Optional]"""
        ui.default.message(
            "If you'd like to set your okta username in the config file, specify the username\n."
            "This is optional.")
        okta_username = self._get_user_input_with_default_value(
            "Okta User Name", default_entry)
        return okta_username

    def _set_okta_password(self, okta_username):
        """Get and validate okta password. [Optional]"""
        ui.default.message(
            "Set Okta Password that will be stored under system keychain")
        self._get_user_input_for_password(
            okta_username, "Okta Pass Word")

    def _get_company_vpn_region(self, default_list : List[str]):
        """ Get Okta AWS App name """
        ui.default.message(
            "Enter the VPN Region Name")
        company_vpn_region = self._get_user_input_with_predefined_list(
            "Please select VPN Region", default_list)
        return company_vpn_region



"""
Licensed under the Apache License, Version 2.0 (the "License");
You may not use this file except in compliance with the License.
You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and* limitations under the License.*
"""
import sys

from . import ui


class LazyVpnExitBase(Exception):
    def __init__(self, message, return_code, result=None):
        """
        :type message: str
        :type return_code: int
        :type result: str
        """
        super().__init__(message, return_code)
        self.message = message
        self.return_code = return_code
        self.result = result

    def handle(self):
        self.handle_message()
        self.handle_result()
        self.exit()

    def handle_message(self):
        if self.message:
            ui.default.info(self.message)

    def handle_result(self):
        if self.result is not None:
            ui.default.result(self.result)

    def exit(self):
        sys.exit(self.return_code)


class LazyVpnExitSuccess(LazyVpnExitBase):
    def __init__(self, message='', return_code=0, result=''):
        super().__init__(message, return_code, result)


class LazyVpnExitError(LazyVpnExitBase):
    def __init__(self, message='ERROR', return_code=1, output=''):
        super().__init__(message, return_code, output)


class LazyVpnExceptionBase(Exception):
    pass


class LazyVpnError(LazyVpnExceptionBase, LazyVpnExitError):
    pass




from subprocess import call

import keyring

from lazyvpn import errors, ui
from lazyvpn.config import Config, LAZY_VPN_KEYCHAIN_SERVICE_NAME


class LazyVpn(object):

    envvar_list = [
        'OKTA_PASSWORD',
        'OKTA_USERNAME',
    ]

    def _run(self):
        # Generate one-time login code using the secret key
        self.handle_action_configure()
        username = self.conf_arg_dict['okta_username']
        password = keyring.get_password(LAZY_VPN_KEYCHAIN_SERVICE_NAME, username)
        company_vpn_region = self.conf_arg_dict['company_vpn_region']
        if self.conf_arg_dict['up']:
            call(['bash', '-c', f'printf "{username}\n{password}\n2"  | /opt/cisco/anyconnect/bin/vpn -s connect "{company_vpn_region}"'])
        elif self.conf_arg_dict['down']:
            call(['bash', '-c', '/opt/cisco/anyconnect/bin/vpn -s disconnect'])

    def __init__(self, ui=ui.cli):
        self.ui = ui
        self.FILE_ROOT = self.ui.HOME
        self._cache = {}

    def handle_action_configure(self):
        # Create/Update config when configure arg set
        if not self.config.action_configure:
            return
        self.config.update_config_file()
        raise errors.LazyVpnExitSuccess()

    def run(self):
        try:
            self._run()
        except errors.LazyVpnExitBase as exc:
            exc.handle()

    def generate_config(self):
        """ generates a new configuration and populates
        various config caches
        """
        self._cache['config'] = config = Config(gac_ui=self.ui)
        conf_arg_dict = config.get_config_dict()
        conf_arg_dict.update(config.get_args().__dict__)
        self._cache['conf_arg_dict'] = conf_arg_dict

        for value in self.envvar_list:
            if self.ui.environ.get(value):
                key = value.lower()
                self.conf_arg_dict[key] = self.ui.environ.get(value)

        return config

    @property
    def config(self):
        if 'config' in self._cache:
            return self._cache['config']
        config = self.generate_config()
        return config

    @property
    def conf_arg_dict(self):
        """
        :rtype: dict
        """
        # noinspection PyUnusedLocal
        config = self.config
        return self._cache['conf_arg_dict']


if __name__ == "__main__":
    lazy_vpn = LazyVpn().run()



"""
Licensed under the Apache License, Version 2.0 (the "License");
You may not use this file except in compliance with the License.
You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and* limitations under the License.*
"""
import builtins
import getpass
import os
import sys


class UserInterface:
    def __init__(self, environ=os.environ, argv=sys.argv):
        self.environ = environ.copy()
        self.environ_bkp = None
        self.argv = argv[:]
        self.argv_bkp = None
        self.args = self.argv[1:]
        with self:
            self.HOME = os.path.expanduser('~')

    def result(self, result):
        """handles output lines
        :type result: str
        """
        raise NotImplementedError()

    def prompt(self, message):
        """handles input's prompt message, but does not ask for input
        :type message: str
        """
        raise NotImplementedError()

    def message(self, message):
        """handles messages meant for user interactions
        :type message: str
        """
        raise NotImplementedError()

    def read_input(self, hidden=False):
        """returns user input
        :rtype: str
        """
        raise NotImplementedError()

    def notify(self, message):
        """handles messages meant for user notifications
        :type message: str
        """
        raise NotImplementedError()

    def input(self, message=None, hidden=False):
        """handles asking for user input, calls prompt() then read_input()
        :type message: str
        :rtype: str
        """
        self.prompt(message)
        return self.read_input(hidden)

    def info(self, message):
        """handles messages meant for info
        :type message: str
        """
        self.notify(message)

    def warning(self, message):
        """handles messages meant for warnings
        :type message: str
        """
        self.notify(message)

    def error(self, message):
        """handles messages meant for errors
        :type message: str
        """
        self.notify(message)

    def __enter__(self):
        self.environ_bkp = os.environ
        self.argv_bkp = sys.argv

        os.environ = self.environ
        sys.argv = sys.argv[:1] + self.args
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.environ = self.environ_bkp
        sys.argv = self.argv_bkp
        self.environ_bkp = None
        self.argv_bkp = None


class CLIUserInterface(UserInterface):
    def result(self, result):
        builtins.print(result, file=sys.stdout)

    def prompt(self, message=None):
        if message is not None:
            builtins.print(message, file=sys.stderr, end='')
            sys.stderr.flush()

    def message(self, message):
        builtins.print(message, file=sys.stderr)

    def read_input(self, hidden=False):
        return getpass.getpass('') if hidden else builtins.input()

    def notify(self, message):
        builtins.print(message, file=sys.stderr)

cli = CLIUserInterface()
default = cli



from setuptools import setup, find_packages

import lazyvpn

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='lazyvpn',
    version=lazyvpn.version,
    install_requires=requirements,
    author='Yuzhong Li',
    author_email='yuzhongl@zillowgroup.com',
    description="A CLI to connect to vpn via cisco anyconnect in one step",
    url='https://github.com/lyzh945/lazyvpn',
    license='Apache License, v2.0',
    packages=find_packages(exclude=('tests', 'docs')),
    test_suite="tests",
    scripts=['bin/lazyvpn', 'bin/lazyvpn.cmd'],
    classifiers=[
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: Apache Software License'
    ]
)
